/*
 * Scratchbox Remote Shell
 * Copyright (c) 2003 Timo Savola <timo.savola@movial.fi>
 *
 * sbrsh is licensed under GPL.
 * You can read the full license at http://www.gnu.org/licenses/gpl.txt
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <syslog.h>
#include <pwd.h>
#include <netdb.h>
#include <signal.h>
#include <errno.h>
#include <pty.h>
#include <utmp.h>
#include <fcntl.h>
#include <getopt.h>
#include <time.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/select.h>
#include <sys/param.h>
#include <netinet/in.h>
#include "daemon.h"
#include "common.h"
#include "protocol.h"
#include "buffer.h"
#include "fakeroot.h"

#define ISSET(fd, set)           ((fd) >= 0 && FD_ISSET((fd), (set)))

#define debug_cmdline            if (debug_file) print_debug_cmdline

#define send_empty(data, ptype)  write_packet((data)->sd, (ptype), NULL, 0)

/* The environment of the process. */
extern char **environ;

/** Debug is enabled if this is set. */
FILE *debug_file = NULL;

static int daemon_pid = -1;

/** Server port number. */
static int port = DEFAULT_PORT;

static char *mount_cmd = DEFAULT_MOUNT_CMD;
static char *umount_cmd = DEFAULT_UMOUNT_CMD;
static char *mount_tab = DEFAULT_MOUNT_TAB;
static char *bind_opt = DEFAULT_BIND_OPT;
static unsigned int mount_expiration = DEFAULT_MOUNT_EXPIRATION;

/** Doubly linked list of mount entries. */
static struct { mount_t *head, *tail; } mounts = { NULL, NULL };

/** The mounts of the handler processes. */
static pid_mounts_t *pid_mounts = NULL;

/**
 * Prints only timestamp and pid to debug log.
 */
static void print_debug_prefix(void)
{
	struct timeval tv;
	struct tm *t;

	gettimeofday(&tv, NULL);
	t = gmtime(&tv.tv_sec);

	fprintf(debug_file,
		"%02d-%02d-%04d %02d:%02d:%02d.%03ld %5d ",
		t->tm_mday, t->tm_mon + 1, 1900 + t->tm_year,
		t->tm_hour, t->tm_min, t->tm_sec, tv.tv_usec / 1000,
		getpid());
}

/**
 * Prints message to debug log.
 */
void print_debug(const char *msg, ...)
{
	va_list arg;

	print_debug_prefix();

	va_start(arg, msg);
	vfprintf(debug_file, msg, arg);
	va_end(arg);

	fprintf(debug_file, "\n");
	fflush(debug_file);
}

/**
 * Prints command line to debug log.
 */
static void print_debug_cmdline(char **argv)
{
	print_debug_prefix();
	fprintf(debug_file, "Executing command:");

	while (*argv) {
		fprintf(debug_file, " %s", *argv++);
	}

	fprintf(debug_file, "\n");
	fflush(debug_file);
}

/**
 * Prints message and errno description (if non-zero) to syslog and sends it to
 * the client (if data is non-null).
 */
static void print_error(handler_t *data,
			const char *progname,
			const char *msg,
			va_list arg,
			int priority)
{
	char str[1024] = { '\0' }, *err;
	int len;

	vsnprintf(str, sizeof (str) - 1, msg, arg);
	len = strlen(str);

	err = strerror(errno);
	if (errno && err && (len + strlen(err) + 3) < sizeof (str)) {
		strcat(str, " (");
		strcat(str, err);
		strcat(str, ")");

		len = strlen(str);
	}

	if (progname) {
		fprintf(stderr, "%s: %s\n", progname, str);
	} else {
		syslog(priority, str);
	}

	debug(priority == LOG_WARNING ? "Warning: %s" : "Error: %s", str);

	if (data) {
		write_packet(data->sd, PTYPE_ERROR, str, len);
	}
}

void send_error(handler_t *data,
		const char *msg,
		...)
{
	va_list arg;

	va_start(arg, msg);
	print_error(data, NULL, msg, arg, LOG_ERR);
	va_end(arg);
}

void send_oom_error(handler_t *data)
{
	send_error(data, "Out of memory");
}

#define oom_error() send_oom_error(NULL)

/**
 * Prints message and errno description to syslog and stderr.
 */
static void error_err(const char *progname,
		      const char *msg,
		      ...)
{
	va_list arg;

	va_start(arg, msg);
	print_error(NULL, progname, msg, arg, LOG_ERR);
	va_end(arg);
}

/**
 * Prints message and errno description to syslog.
 */
void error(const char *msg,
	   ...)
{
	va_list arg;

	va_start(arg, msg);
	print_error(NULL, NULL, msg, arg, LOG_ERR);
	va_end(arg);
}

/**
 * Prints message to syslog.
 */
static void warn(const char *msg,
		 ...)
{
	va_list arg;

	errno = 0;

	va_start(arg, msg);
	print_error(NULL, NULL, msg, arg, LOG_WARNING);
	va_end(arg);
}

/**
 * Sends message to the client.
 */
static void send_message(handler_t *data,
			 const char *msg)
{
	debug("Sending message: %s", msg);
	write_packet(data->sd, PTYPE_MESSAGE, (char *) msg, strlen(msg));
}

/**
 * Passes the contents of BUF to send_error().
 */
static void flush(handler_t *data,
		  char *buf,
		  int *lenp)
{
	if (*lenp > 0) {
		buf[*lenp] = '\0';
		*lenp = 0;

		send_error(data, buf);
	}
}

/**
 * Executes argv[0] in a child process.
 * @return -1 on error, 0 otherwise
 */
static int execute(handler_t *data,
		   char **argv)
{
	int pid, status, err[2], pos = 0;
	char line[1024];

	if (pipe(err) < 0) {
		send_error(data, "Can't create pipe");
		return -1;
	}

	pid = fork();
	if (pid < 0) {
		return -1;
	}

	if (pid == 0) {
		/* Child */

		if (dup2(err[1], STDOUT_FILENO) != STDOUT_FILENO) {
			send_error(data, "Can't duplicate pipe as stdout");
			exit(1);
		}
		if (dup2(err[1], STDERR_FILENO) != STDERR_FILENO) {
			send_error(data, "Can't duplicate pipe as stderr");
			exit(1);
		}
		close(err[0]);

		execv(argv[0], argv);

		send_error(data, "Can't execute command: %s", argv[0]);
		exit(1);
	}

	/* Parent */

	close(err[1]);

	while (1) {
		char c;

		if (read(err[0], &c, 1) <= 0) {
			flush(data, line, &pos);
			break;
		}

		if (c == '\n' || c == '\0') {
			flush(data, line, &pos);
			continue;
		}

		if (pos == sizeof (line)) {
			flush(data, line, &pos);
		}

		line[pos++] = c;
	}

	if (waitpid(pid, &status, 0) != pid) {
		return -1;
	}

	if (status < 0) {
		return -1;
	}

	errno = 0;
	return (WEXITSTATUS(status) != 0) ? -1 : 0;
}

/**
 * Finds a host from a config file and returns the associated password.
 *
 * Example config file entry:
 * 192.168.0.2  foobar
 *
 * @param filename the config file path
 * @return the password if found, NULL if not
 */
static char *find_pwd(handler_t *data,
		      const char *filename)
{
	FILE *file;
	char buf[1024], *buf2, *pwd = NULL;
	int len;

	file = fopen(filename, "r");
	if (!file) {
		send_error(data, "Can't open config file: %s", filename);
		return NULL;
	}

	while (1) {
		len = read_line(file, buf, sizeof (buf));
		if (len < 0) {
			break;
		}
		if (len == 0 || isspace(buf[0])) {
			continue;
		}

		buf2 = find_space(buf);
		if (*buf2 == '\0') {
			continue;
		}
		*buf2++ = '\0';

		if (strcmp(buf, data->host) != 0) {
			continue;
		}

		buf2 = skip_spaces(buf2);
		if (*buf2 == '\0') {
			continue;
		}
		*find_space(buf2) = '\0';

		pwd = strdup(buf2);
		if (!pwd) {
			send_oom_error(data);
		}
		break;
	}

	fclose(file);

	return pwd;
}

/**
 * Recursively creates all needed directories (with MKDIR_PERMS).
 */
static int mkdirs(const char *path)
{
	char *parent, *p;
	int rc = -1;

	if (mkdir(path, MKDIR_PERMS) == 0 || errno == EEXIST) {
		return 0;
	}

	if (errno != ENOENT) {
		return -1;
	}

	parent = strdup(path);
	if (!parent) {
		return -1;
	}

	p = strrchr(parent, '/');
	if (p && parent != p) {
		*p = '\0';
		rc = 0;
	}

	if (rc == 0) {
		rc = mkdirs(parent);
	}

	free(parent);

	if (rc == 0) {
		rc = mkdir(path, MKDIR_PERMS);
	}

	return rc;
}

/**
 * Frees all resources associated with a mount entry.
 */
static void mnt_free(mount_t *m)
{
	if (m->opts) {
		free(m->opts);
	}

	if (m->device) {
		free(m->device);
	}

	if (m->point) {
		free(m->point);
	}

	free(m);
}

/**
 * Deep-copies a mount entry.
 */
static mount_t *mnt_copy(mount_t *mnt)
{
	mount_t *c;

	c = calloc(1, sizeof (mount_t));
	if (!c) {
		return NULL;
	}

	c->type = mnt->type;

	c->point = strdup(mnt->point);
	if (!c->point) {
		goto _error;
	}

	c->device = strdup(mnt->device);
	if (!c->device) {
		goto _error;
	}

	if (mnt->opts) {
		c->opts = strdup(mnt->opts);
		if (!c->opts) {
			goto _error;
		}
	}

	c->device_dev = mnt->device_dev;
	c->point_dev = mnt->point_dev;

	return c;

_error:
	mnt_free(mnt);
	return NULL;
}

/**
 * Parses a string of form: <type> <device> <point> [<options>] [<device num>]
 */
static mount_t *mnt_parse(handler_t *data,
			  char *str,
			  const char *rootdir)
{
	char *typestr, *dev, *pnt, *opt, *devnum;
	int type;
	mount_t *m;

	split_string(str, &typestr, &dev, &pnt, &opt, &devnum, NULL);

	if (!typestr) {
		send_error(data, "No type in mount string");
		return NULL;
	}
	if (!dev) {
		send_error(data, "No device in mount string");
		return NULL;
	}
	if (!pnt) {
		send_error(data, "No mount point in mount string");
		return NULL;
	}

	if (strcmp(typestr, MTYPE_NFS_S) == 0) {
		type = MTYPE_NFS;
	} else if (strcmp(typestr, MTYPE_BIND_S) == 0) {
		type = MTYPE_BIND;
	} else {
		send_error(data, "Invalid mount type \"%s\"", typestr);
		return NULL;
	}

	m = calloc(sizeof (mount_t), 1);
	if (!m) {
		send_oom_error(data);
		return NULL;
	}

	m->type = type;

	m->device = strdup(dev);
	if (!m->device) {
		goto _oom_free;
	}

	m->point = malloc(strlen(rootdir) + strlen(pnt) + 1);
	if (!m->point) {
		goto _oom_free;
	}
	strcpy(m->point, rootdir);
	strcat(m->point, pnt);

	/* Remove / from the end. */
	str = strrchr(m->point, '/');
	if (str && strlen(str) == 1) {
		*str = '\0';
	}

	m->opts = NULL;
	if (opt && strcmp(opt, "-") != 0) {
		m->opts = strdup(opt);
		if (!m->opts) {
			goto _oom_free;
		}
	}

	m->device_dev = 0;
	if (devnum) {
		char *end;

		m->device_dev = strtoll(devnum, &end, 16);

		if (*end != 'h') {
			send_error(data, "Invalid device number: %s", devnum);
			goto _free;
		}
	}

	m->point_dev = 0;

	return m;

_oom_free:
	send_oom_error(data);
_free:
	mnt_free(m);
	return NULL;
}

/**
 * Finds an entry from the MOUNTS list.
 * @param point the mount point of the entry
 */
static mount_t *mnt_list_find(const char *point)
{
	mount_t *mnt;

	for (mnt = mounts.head; mnt; mnt = mnt->next) {
		if (strcmp(mnt->point, point) == 0) {
			break;
		}
	}

	return mnt;
}

/**
 * Adds an entry to the MOUNTS list. The entries are kept in ascending order
 * based on their mount point strings.
 */
static void mnt_list_add(mount_t *mnt)
{
	mount_t *prev, *next;

	prev = NULL;
	next = mounts.head;
	while (next) {
		if (strcmp(next->point, mnt->point) > 0) {
			break;
		}

		prev = next;
		next = next->next;
	}

	mnt->prev = prev;
	mnt->next = next;

	if (prev) {
		prev->next = mnt;
	} else {
		mounts.head = mnt;
	}

	if (next) {
		next->prev = mnt;
	} else {
		mounts.tail = mnt;
	}
}

/**
 * Removes and frees an entry from the MOUNTS list.
 */
static void mnt_list_del(mount_t *mnt)
{
	if (mnt->prev) {
		mnt->prev->next = mnt->next;
	} else {
		mounts.head = mnt->next;
	}

	if (mnt->next) {
		mnt->next->prev = mnt->prev;
	} else {
		mounts.tail = mnt->prev;
	}

	mnt_free(mnt);
}

/**
 * Executes mount_cmd.
 */
static int mnt_mount(handler_t *data,
		     const mount_t *m)
{
	char *argv[8];

	memset(argv, 0, sizeof (argv));

	argv[0] = mount_cmd;
	argv[1] = m->device;
	argv[2] = m->point;

	switch (m->type) {
	case MTYPE_NFS:
		argv[3] = "-t";
		argv[4] = "nfs";
		argv[5] = m->opts ? "-o" : NULL;
		argv[6] = m->opts;

		if (m->opts) {
			debug("Executing: %s %s %s -t nfs -o %s",
			      argv[0], argv[1], argv[2], argv[6]);
		} else {
			debug("Executing: %s %s %s -t nfs",
			      argv[0], argv[1], argv[2]);
		}

		break;

	case MTYPE_BIND:
		split_string(bind_opt,
			     &argv[3], &argv[4], &argv[5], &argv[6], NULL);

		debug("Executing: %s %s %s %s %s %s %s",
		      argv[0], argv[1], argv[2],
		      argv[3] ? argv[3] : "",
		      argv[4] ? argv[4] : "",
		      argv[5] ? argv[5] : "",
		      argv[6] ? argv[6] : "");

		break;

	default:
		errno = EINVAL;
		return -1;
	}

	return execute(data, argv);
}

static int unmount(handler_t *data,
		   char *point)
{
	char *argv[] = { umount_cmd, point, NULL };

	debug("Executing: %s %s", argv[0], argv[1]);

	return execute(data, argv);
}

/**
 * Checks /etc/mtab if m->point is already mounted.
 * @param oldp will be set if already mounted
 * @return -1 on error, 0 if wasn't mounted, 1 if was mounted
 */
static int mnt_touch(handler_t *data,
		     mount_t *mnt,
		     mount_t **oldp)
{
	FILE *file;
	char buf[1024], *dev, *pnt, *type;
	int len, mounted = 0;
	mount_t *old;

	/* Check the file. */

	file = fopen(mount_tab, "r");
	if (!file) {
		send_error(data, "Can't open mount table file: %s", mount_tab);
		return -1;
	}

	while (1) {
		len = read_line(file, buf, sizeof (buf));
		if (len < 0) {
			break;
		}

		split_string(buf, &dev, &pnt, &type, NULL);
		if (!dev || !pnt) {
			continue;
		}

		if (strcmp(pnt, mnt->point) == 0) {
			mounted = 1;
			break;
		}
	}

	fclose(file);

	/* Check the list. */

	old = mnt_list_find(mnt->point);

	if (mounted) {
		debug("%s is mounted (had %d users)", mnt->point,
		      old ? old->usage : 0);

		if (old) {
			old->usage++;
		} else {
			old = mnt_copy(mnt);
			if (!old) {
				return -1;
			}

			old->usage = 1;
			mnt_list_add(old);
		}

		*oldp = old;
	} else if (old) {
		debug("%s is not mounted but exists on our list with %d users",
		      mnt->point, mnt->usage);

		mnt_list_del(old);
	}

	return mounted;
}

/**
 * Mounts a filesystem if not already mounted.
 */
static mount_t *add_mount(handler_t *data,
			  mount_t *mnt,
			  int uid,
			  int gid)
{
	mount_t *old = NULL, *canon;
	int mounted;

	mounted = mnt_touch(data, mnt, &old);
	if (mounted < 0) {
		return NULL;
	}

	if (old) {
		canon = old;
	} else {
		canon = mnt_copy(mnt);
	}

	if (!mounted) {
		debug("Changing effective uid to %d", gid);

		if (seteuid(uid) < 0) {
			send_error(data, "Can't change effective uid to %d",
				   uid);
			return NULL;
		}

		debug("Creating directory %s", canon->point);

		if (mkdirs(canon->point) < 0) {
			send_error(data, "Can't create mount point: %s",
				   canon->point);
			return NULL;
		}

		debug("Changing effective uid back to root");

		if (seteuid(0) < 0) {
			send_error(data, "Can't change effective uid to 0");
			return NULL;
		}

		if (mnt_mount(data, canon) < 0) {
			return NULL;
		}

		canon->usage = 1;
		mnt_list_add(canon);
	}

	return canon;
}

/**
 * Unmounts a filesystem.
 * @return -1 on error, 0 otherwise
 */
static int remove_mount(handler_t *data,
			char *str,
			const char *rootdir)
{
	mount_t *mnt, *old = NULL;
	int mounted;

	mnt = mnt_parse(data, str, rootdir);
	if (!mnt) {
		return -1;
	}

	mounted = mnt_touch(data, mnt, &old);

	if (mounted >= 0 && old) {
		mnt_list_del(old);
	}

	if (mounted != 0) {
		unmount(data, mnt->point);
	}

	mnt_free(mnt);

	return (mounted < 0) ? -1 : 0;
}

/**
 * Decrements the usage count of a mount and possibly sets its expiration time.
 */
static void release_mount(mount_t *mnt)
{
	mnt->usage--;

	debug("Releasing mount %s (%d users remain)", mnt->point, mnt->usage);

	if (mnt->usage <= 0) {
		mnt->expiration = time(NULL) + mount_expiration;
	}
}

/**
 * Steps through the MOUNTS list and unmounts all unused and expired entries.
 */
static void expire_mounts(void)
{
	mount_t *prev, *curr;
	time_t now;

	debug("Checking for expired mounts");

	now = time(NULL);

	prev = NULL;
	curr = mounts.tail;
	while (curr) {
		prev = curr->prev;

		if (curr->usage <= 0 &&
		    curr->expiration <= now) {
			debug("Unmounting %s", curr->point);

			if (curr->next) {
				curr->next->prev = curr->prev;
			} else {
				mounts.tail = curr->prev;
			}

			if (curr->prev) {
				curr->prev->next = curr->next;
			} else {
				mounts.head = curr->next;
			}

			unmount(NULL, curr->point);
			mnt_free(curr);
		}

		curr = prev;
	}
}

/**
 * Unmounts all entries in the MOUNTS list.
 */
static void unmount_all(void)
{
	mount_t *prev, *mnt;

	mnt = mounts.tail;
	while (mnt) {
		prev = mnt->prev;

		debug("Unmounting %s", mnt->point);

		unmount(NULL, mnt->point);
		mnt_free(mnt);

		mnt = prev;
	}

	mounts.head = NULL;
	mounts.tail = NULL;
}

/**
 * Gets the amount of data we can read from a file (which may be a tty).
 */
static int get_data_length(handler_t *data,
			   int fd,
			   int max)
{
	int len;

	len = max;

	if (isatty(fd)) {
		if (ioctl(fd, FIONREAD, &len) < 0) {
			send_error(data, "Can't check pty for available data");
			return -1;
		}

		if (len < 0) {
			send_error(data, "ioctl() gave invalid read length");
			return -1;
		}

		if (len > max) {
			len = max;
		}
	}

	return len;
}

/**
 * Reads some data from a fd and writes it into the socket (sd) in a packet.
 * The fd will be set to -1 at EOF.
 * @param fd pointer to the stdout or stderr of the child process
 * @param ptype the packet type
 */
static void send_data(handler_t *data,
		     output_desc_t *out)
{
	int len;

	len = get_data_length(data, out->fd, out->req);
	if (len < 0) {
		goto _error;
	}

	if (len == 0) {
		/* Nothing to do. */
		return;
	}

	len = read(out->fd, data->tmp_buf, len);
	if (len < 0) {
		send_error(data, "Can't read output of the process");
		goto _error;
	}

	if (len) {
		debug("Sending %s DATA packet (%d bytes)",
		      output_desc_is_stdout(out) ? "OUT" : "ERR", len);

		if (write_packet(data->sd, out->data_ptype,
				 data->tmp_buf, len) < 0) {
			error("Can't write packet to socket");
			goto _error;
		}

		out->req -= len;
	} else {
		debug(output_desc_is_stdout(out) ?
		      "Stdout hit EOF" : "Stderr hit EOF");

		out->fd = -1;
		out->req = 0;
	}

	return;

_error:
	data->error = 1;
}

/**
 * Writes an RC or AUTH OK packet.
 */
static int send_int16(handler_t *data,
		      int ptype,
		      int16_t val)
{
	val = htons(val);

	if (write_packet(data->sd, ptype, &val, sizeof (val)) < 0) {
		error("Can't write int16 packet to socket");
		return -1;
	}

	return 0;
}

/**
 * Sends a REQuest for more DATA.
 */
static void send_request(handler_t *data,
			input_desc_t *in)
{
	/* Already waiting? */
	if (in->wait) {
		return;
	}

	debug("Sending IN REQ packet");

	if (write_packet(data->sd, in->req_ptype, NULL, 0) < 0) {
		error("Can't write packet to socket");
		goto _error;
	}

	in->wait = 1;
	return;

_error:
	data->error = 1;
}

/**
 * Writes (some of) in->buf to in->fd.
 */
static void write_buffer(handler_t *data,
			 input_desc_t *in)
{
	debug("Writing buffer to stdin");

	if (buf_write_out(&in->buf, &in->fd) < 0) {
		debug("Can't write to stdin (%s)", strerror(errno));
	}

	if (buf_is_empty(&in->buf)) {
		in->wait = 0;
	}
}

/**
 * Reads bytes from socket to buf_in.
 * @param len the maximum number of bytes to be read
 */
static void receive_stream(handler_t *data,
			   int len)
{
	if (len > 0) {
		debug("Receiving IN DATA packet (%d bytes)", len);

		if (buf_read_in(&data->in.buf, data->sd, len) < 0) {
			send_error(data,"Can't read IN DATA packet to buffer");
			goto _error;
		}
	} else {
		debug("Receiving IN DATA packet: EOF");

		buf_set_eof(&data->in.buf);
	}

	return;

_error:
	data->error = 1;
}

/**
 * Read an IN DATA or OUT/ERR REQ packet from socket.
 */
static void receive_packet(handler_t *data)
{
	phead_t head;

	if (read_phead(data->sd, &head) < 0) {
		error("Can't read packet header from socket");
		goto _error;
	}

	switch (head.type) {
	case PTYPE_IN_DATA:
		receive_stream(data, head.size);
		return;

	case PTYPE_OUT_REQ:
		debug("Receiving OUT REQ packet");

		if (head.size) {
			send_error(data, "OUT REQ packet size was non-zero");
			goto _error;
		}

		data->out.req = BUFFER_SIZE;
		return;

	case PTYPE_ERR_REQ:
		debug("Receiving ERR REQ packet");

		if (head.size) {
			send_error(data, "ERR REQ packet size was non-zero");
			goto _error;
		}

		data->err.req = BUFFER_SIZE;
		return;
	}

	send_error(data, "Packet has invalid type: %d", head.type);

_error:
	data->error = 1;
}

/**
 * Prints information on an exited process (pid) to debug log.
 */
static void print_status(int pid,
			 int status)
{
	int sig;

	if (WIFEXITED(status)) {
		debug("Process %d returned: %d", pid, WEXITSTATUS(status));
	} else if (WIFSIGNALED(status)) {
		sig = WTERMSIG(status);
		debug("Process %d terminated by signal: %s (%d)",
		      pid, strsignal(sig), sig);
	} else if (WIFSTOPPED(status)) {
		sig = WSTOPSIG(status);
		debug("Process %d stopped by signal: %s (%d)",
		      pid, strsignal(sig), sig);
	} else {
		error("Invalid status %d for pid %d", status, pid);
	}
}

#define print_debug_bool(m, b)  print_debug(m, (b) ? "yes" : "no")
#define debug_bool(m, b)        if (debug_file) print_debug_bool(m, b)

#define debug_status() \
	if (debug_file) { \
		print_debug_bool("Process alive  = %s", alive); \
		print_debug_bool("Stdin open     = %s", data->in.fd >= 0); \
		print_debug_bool("Stdin waiting  = %s", data->in.wait); \
		print_debug     ("Stdin buffered = %d bytes%s", \
				 buf_size(&data->in.buf), \
				 data->in.buf.eof ? " (EOF set)" : ""); \
		print_debug_bool("Stdout open    = %s", data->out.fd >= 0); \
		print_debug     ("Stdout request = %d bytes", data->out.req); \
		print_debug_bool("Stderr open    = %s", data->err.fd >= 0); \
		print_debug     ("Stderr request = %d bytes", data->err.req); \
		print_debug_bool("Error          = %s", data->error); \
	}

#define debug_select(msg) \
	if (debug_file) { \
		print_debug(msg " [%3s %2s %3s %3s][%2s]", \
			    ISSET(data->sd,      &readfds) ? "NET" : "", \
			    ISSET(data->in.fd,   &readfds) ? "IN"  : "", \
			    ISSET(data->out.fd,  &readfds) ? "OUT" : "", \
			    ISSET(data->err.fd,  &readfds) ? "ERR" : "", \
			    ISSET(data->in.fd,  &writefds) ? "IN"  : ""); \
	}

/**
 * Manages the streams of the command executor (= the child-child process).
 * @param pid of the child process
 * @return the return code (-1 on error)
 */
static int16_t handler_manage(handler_t *data,
			      pid_t pid)
{
	fd_set readfds, writefds;
	int maxfd1, status, alive = 1;

	debug("Managing process %d", pid);

	maxfd1 = MAX(data->sd, data->in.fd);
	maxfd1 = MAX(maxfd1, data->out.fd);
	maxfd1 = MAX(maxfd1, data->err.fd);
	maxfd1++;

	if (!buf_init(&data->in.buf)) {
		send_oom_error(data);
		goto _kill;
	}

	while (1) {
		int count, val;

		debug_status();

		if (data->error) {
			goto _kill;
		}

		/* Non-TTY mode: Check if we should exit. */
		if (!data->cmd.is_tty &&
		    data->in.fd < 0 &&
		    data->out.fd < 0 &&
		    data->err.fd < 0) {
			break;
		}

		FD_ZERO(&readfds);
		FD_ZERO(&writefds);

		FD_SET(data->sd, &readfds);

		if (data->out.fd >= 0 && data->out.req > 0) {
			FD_SET(data->out.fd, &readfds);
		}

		if (!data->cmd.is_tty &&
		    data->err.fd >= 0 && data->err.req > 0) {
			FD_SET(data->err.fd, &readfds);
		}

		if (data->in.fd >= 0) {
			if (!data->cmd.is_tty) {
				/* We want select to wake up if stdin hits EOF
				   (child exited but we haven't noticed yet) */
				FD_SET(data->in.fd, &readfds);
			}

			if (!buf_is_empty(&data->in.buf) || !data->in.wait) {
				FD_SET(data->in.fd, &writefds);
			}
		}

		debug_select("Selecting");

		count = select(maxfd1, &readfds, &writefds, NULL, NULL);
		if (count < 0) {
			if (errno == EINTR) {
				debug("Select interrupted");
			} else {
				error("Select failed");
				goto _kill;
			}
		} else if (count > 0) {
			debug_select("Selected ");

			/* TTY mode: Check if we should exit. */
			if (data->cmd.is_tty &&
			    !alive &&
			    ISSET(data->out.fd, &readfds) &&
			    get_data_length(data, data->out.fd, 1) <= 0) {
				break;
			}

			if (ISSET(data->sd, &readfds)) {
				receive_packet(data);
			}

			if (ISSET(data->out.fd, &readfds)) {
				send_data(data, &data->out);
			}

			if (ISSET(data->err.fd, &readfds)) {
				send_data(data, &data->err);
			}

			if (ISSET(data->in.fd, &writefds)) {
				if (buf_is_empty(&data->in.buf)) {
					/* Request data to buffer. */
					send_request(data, &data->in);
				} else {
					/* Flush buffer. */
					write_buffer(data, &data->in);
				}
			}

			/* Stdin at EOF? */
			if (!data->cmd.is_tty &&
			    ISSET(data->in.fd, &readfds)) {
				data->in.fd = -1;
			}
		}

		/* Collect late children. */

		val = waitpid(-1, &status, WNOHANG);
		if (val < 0 && errno != ECHILD) {
			send_error(data, "Can't wait for children");
			return -1;
		}

		if (val > 0) {
			if (val == pid) {
				alive = 0;
			}

			print_status(val, status);
		}
	}

	if (alive) {
		if (waitpid(pid, &status, 0) < 0) {
			send_error(data, "Can't wait for child %d", pid);
			return -1;
		}

		print_status(pid, status);
	}

	if (WIFEXITED(status)) {
		return WEXITSTATUS(status);
	}
	return -1;

_kill:
	kill(pid, SIGTERM);
	return -1;
}

/**
 * @return the truth or -1 on error
 */
static int authenticate(handler_t *data,
			const char *home)
{
	char *path, *pwd;
	int ok;

	if (!home) {
		send_error(data, "No home directory found");
		return -1;
	}

	path = malloc(strlen(home) + strlen("/" CONFIG_NAME) + 1);
	if (!path) {
		send_oom_error(data);
		return -1;
	}
	strcpy(path, home);
	strcat(path, "/" CONFIG_NAME);

	debug("Searching for password in %s", path);

	pwd = find_pwd(data, path);

	free(path);

	if (!pwd) {
		warn("Can't find password from config file");
		return 0;
	}

	ok = (strcmp(pwd, data->auth.pwd) == 0);

	free(pwd);

	if (!ok) {
		warn("Passwords don't match");
	}

	return ok;
}

/**
 * More or less like forkpty.
 * @return pid of the child process (0 for child) or -1 on error
 */
static int fork_pty(handler_t *data,
		    int *ptyfd)
{
	int master, slave, pid;
	struct winsize ws;

	ws.ws_row = data->cmd.tty_row;
	ws.ws_col = data->cmd.tty_col;
	ws.ws_xpixel = data->cmd.tty_xpixel;
	ws.ws_ypixel = data->cmd.tty_ypixel;

	if (openpty(&master, &slave, NULL, NULL, &ws) < 0) {
		send_error(data, "Can't open a pseudo-tty");
		return -1;
	}

	pid = fork();
	if (pid < 0) {
		send_error(data, "Can't fork");
		return -1;
	}

	if (pid == 0) {
		/* Child */

		close(master);
		if (login_tty(slave)) {
			send_error(data, "Can't login to a pseudo-tty");
			exit(1);
		}
	} else {
		/* Parent */

		close(slave);
		*ptyfd = master;
	}

	return pid;
}

/**
 * More or less like fork, but the stdin/stdout/stderr of the child process
 * are set to point to sockets.
 * @param infd a place to write the parent-end of the child's stdin socketpair
 * @return pid of the child process (0 for child) or -1 on error
 */
static int fork_sockets(handler_t *data,
			int *infd,
			int *outfd,
			int *errfd)
{
	int insd[2], outsd[2], errsd[2], pid;

	if (socketpair(AF_UNIX, SOCK_STREAM, 0, insd) < 0 ||
	    socketpair(AF_UNIX, SOCK_STREAM, 0, outsd) < 0 ||
	    socketpair(AF_UNIX, SOCK_STREAM, 0, errsd) < 0) {
		send_error(data, "Can't create socket pairs");
		return -1;
	}

	pid = fork();
	if (pid < 0) {
		send_error(data, "Can't fork");
		return -1;
	}

	if (pid == 0) {
		/* Child */

		close(insd[0]);
		close(outsd[0]);
		close(errsd[0]);

		if (dup2(insd[1], STDIN_FILENO) != STDIN_FILENO) {
			send_error(data, "Can't duplicate socket as stdin");
			exit(1);
		}

		if (dup2(outsd[1], STDOUT_FILENO) != STDOUT_FILENO) {
			send_error(data, "Can't duplicate socket as stdout");
			exit(1);
		}

		if (dup2(errsd[1], STDERR_FILENO) != STDERR_FILENO) {
			send_error(data, "Can't duplicate socket as stderr");
			exit(1);
		}
	} else {
		/* Parent */

		if (set_nonblocking(insd[0])) {
			send_error(data, "Can't make socket non-blocking");
			return -1;
		}

		*infd = insd[0];
		*outfd = outsd[0];
		*errfd = errsd[0];
	}

	close(insd[1]);
	close(outsd[1]);
	close(errsd[1]);

	return pid;
}

/**
 * Unmounts all filesystems listed in command info's option vector.
 * @return 0 or -1 on error
 */
static int unmount_opts(handler_t *data)
{
	char **optp;
	int cnt, error = 0;

	debug("Unmounting filesystems");

	/* Count the options and go to end of the vector. */
	for (cnt = 0, optp = data->cmd.opts; *optp; cnt++, optp++);

	/* Try to unmount filesystems in reverse order. */
	for (optp--; cnt-- > 0; optp--) {
		if (remove_mount(data, *optp, data->root) < 0) {
			error = -1;
		}
	}

	return error;
}

/**
 * Bubble sort based on mount point string.
 */
static void sort_mounts(mount_t **mounts)
{
	mount_t **p;
	int touched;

	if (!mounts[0] || !mounts[1]) {
		return;
	}

	do {
		touched = 0;

		for (p = mounts; p[1]; p++) {
			if (strcmp(p[0]->point, p[1]->point) > 0) {
				mount_t *tmp;

				tmp = p[0];
				p[0] = p[1];
				p[1] = tmp;

				touched = 1;
			}
		}
	} while (touched);
}

/**
 * Read "SBRSH_RLIMIT"-fields from ENVIRONment and set them in place.
 * @return -1 on error, 0 otherwise
 */
static int set_rlimits(handler_t *data)
{
	static struct { const char *key; int resource; } info[] = {
		{ "SBRSH_RLIMIT_CPU",		RLIMIT_CPU	},
		{ "SBRSH_RLIMIT_FSIZE",		RLIMIT_FSIZE	},
		{ "SBRSH_RLIMIT_DATA",		RLIMIT_DATA	},
		{ "SBRSH_RLIMIT_STACK",		RLIMIT_STACK	},
		{ "SBRSH_RLIMIT_CORE",		RLIMIT_CORE	},
		{ "SBRSH_RLIMIT_RSS",		RLIMIT_RSS	},
		{ "SBRSH_RLIMIT_NPROC",		RLIMIT_NPROC	},
		{ "SBRSH_RLIMIT_NOFILE",	RLIMIT_NOFILE	},
		{ "SBRSH_RLIMIT_MEMLOCK",	RLIMIT_MEMLOCK	},
		{ "SBRSH_RLIMIT_AS",		RLIMIT_AS	},
		{ NULL,				0		}
	};

	int i;
	char *str;
	struct rlimit lim;

	for (i = 0; info[i].key; i++) {
		str = getenv(info[i].key);
		if (!str) {
			continue;
		}

		str = skip_spaces(str);
		if (strlen(str) == 0) {
			continue;
		}

		debug("Setting %s to %s", info[i].key + 6, str);

		if (getrlimit(info[i].resource, &lim) < 0) {
			send_error(data, "Can't get %s resource limit",
				   info[i].key + 13);
			return -1;
		}

		lim.rlim_cur = RLIM_INFINITY;

		if (strncmp(str, ENV_RLIMIT_UNLIMITED,
			    strlen(ENV_RLIMIT_UNLIMITED)) != 0) {
			lim.rlim_cur = atol(str);

			if (lim.rlim_cur == 0 && strcmp(str, "0") != 0) {
				send_error(data, "Invalid %s resource limit "
					   "value: %s", info[i].key + 13, str);
				return -1;
			}
		}

		if (setrlimit(info[i].resource, &lim) < 0) {
			send_error(data, "Can't set %s resource limit to %d"
				   " while maximum is %d", info[i].key + 13,
				   lim.rlim_cur, lim.rlim_max);
			return -1;
		}
	}

	return 0;
}

static void modify_environment_vector(char **src,
				      char *entry,
				      char *newentry)
{
	char **dest, *key, *value, len = 0;

	key = newentry;

	/* Length of the key (including '='). */
	value = strchr(key, '=');
	if (value) {
		value++;
		len = value - key;
	}

	/* Copy the vector onto itself skipping ENTRY and the key that
	 * equals KEY. */
	for (dest = src; *src; src++) {
		if (*src != entry && strncmp(*src, key, len) != 0) {
			*dest++ = *src;
		}
	}

	/* Put NEWENTRY at the end of the vector if
	 * its value is not "(UNSET)". */
	if (value && strcmp(value, ENV_RENAME_UNSET) != 0) {
		*dest++ = newentry;
	}

	/* Nul-terminate the vector (which has a nul-terminator at its
	 * old end). */
	while (*dest) {
		*dest++ = NULL;
	}
}

static void set_environment(char **env)
{
	char **p, *entry;
	const int prelen = strlen(ENV_RENAME_PREFIX);

	/* Strip "SBRSH_ENV_" prefices. (This is a bit dirty, but doesn't
	 * matter since we'll exec after this.) */

	while (1) {
		entry = NULL;

		/* Find first SBRSH_ENV-entry. */
		for (p = env; *p; p++) {
			entry = *p;
			if (strncmp(entry, ENV_RENAME_PREFIX, prelen) == 0) {
				break;
			} else {
				entry = NULL;
			}
		}

		/* Quit if there are no entries left. */
		if (!entry) {
			break;
		}

		debug("Converting \"%s\"", entry);
		modify_environment_vector(env, entry, entry + prelen);
	}

	/* Hit it! */
	environ = env;
}

/**
 * Copies a string and adds a "/" prefix if it doesn't have one.
 * abort()s if ENOMEM. This function is evil!
 */
static char *add_root_prefix(handler_t *data,
			     char *orig)
{
	char *dir;

	if (orig[0] == '/') {
		return orig;
	}

	dir = malloc(1 + strlen(orig) + 1);
	if (!dir) {
		send_oom_error(data);
		exit(1);
	}

	strcpy(dir, "/");
	strcat(dir, orig);

	return dir;
}

/**
 * Executes /The/ command. Never returns. Dies with abort() so that cleanup()
 * is not called.
 */
static void execute_command(handler_t *data)
{
	char *dir;
	struct passwd *pass;
	char **argv;

	set_closeonexec(data->sd);

	dir = add_root_prefix(data, data->root);
	debug("Changing root directory to %s", dir);
	if (chroot(dir) < 0) {
		send_error(data, "Can't change root directory to: %s", dir);
		goto _error;
	}

	dir = add_root_prefix(data, data->cmd.curdir);
	debug("Changing current directory to %s", dir);
	if (chdir(dir) < 0) {
		send_error(data, "Can't change current directory to: %s "
			   "[after chroot]", dir);
		goto _error;
	}

	debug("Getting passwd struct of user %s (after chroot)",
	      data->auth.user);
	pass = getpwnam(data->auth.user);
	if (!pass) {
		send_error(data, "getpwnam: Can't get information about "
			   "user: %s [after chroot]", data->auth.user);
		goto _error;
	}

	/* Change gid before uid because we need to be root. */

	debug("Changing gid to %d", pass->pw_gid);
	if (setgid(pass->pw_gid) < 0) {
		send_error(data, "Can't change group id to %d [after chroot]",
			   pass->pw_gid);
		goto _error;
	}

	debug("Changing uid to %d", pass->pw_uid);
	if (setuid(pass->pw_uid) < 0) {
		send_error(data, "Can't change user id to %d [after chroot]",
			   pass->pw_uid);
		goto _error;
	}

	/* Change environment and resource limits. */

	set_environment(data->cmd.env);

	if (set_rlimits(data) < 0) {
		goto _error;
	}

	/* Run a fakeroot relay. */

	if (fakeroot_relay(data) < 0) {
		goto _error;
	}

	/* Command and arguments. */

	argv = data->cmd.argv;

	if (argv[0] == NULL) {
		argv = calloc(2, sizeof (char *));
		if (!argv) {
			send_oom_error(data);
			goto _error;
		}

		argv[0] = pass->pw_shell;
		if (argv[0] == NULL) {
			send_error(data, "User %s has no shell",
				   data->auth.user);
			goto _error;
		}
	}

	debug_cmdline(argv);

	/* Do it! */

	execvp(argv[0], argv);

	send_error(data, "Can't execute command: %s", argv[0]);

_error:
	exit(1);
}

/**
 * Does the handshake & stuff for a client.
 * @return the mounts of this handler; NULL on error
 */
static mount_t **handler_startup(handler_t *data)
{
	char strbak[512];
	struct passwd *pass = NULL;
	int ok;
	char **optp;
	mount_t **mntp, **mounts = NULL;

	debug("Client IP is %s", data->host);

	debug("Sending VERSION packet");

	if (send_version(data->sd) < 0) {
		error("Can't send protocol version packet");
		goto _error;
	}

	debug("Reading client's VERSION packet");

	data->client_version = get_version(data->sd);
	if (data->client_version < 0) {
		error("Can't read protocol version packet");
		goto _error;
	}

	if (data->client_version != PROTOCOL_VERSION) {
		warn("Client uses %s protocol version %d",
		     data->client_version < PROTOCOL_VERSION ?
		     "older" : "newer", data->client_version);
	} else {
		debug("Protocol versions match");
	}

	debug("Reading auth info struct from socket");

	if (read_auth(data->sd, &data->auth) < 0) {
		error("Can't read authentication info");
		goto _error;
	}

	debug("Getting passwd struct of user %s (before chroot)",
	      data->auth.user);

	pass = getpwnam(data->auth.user);
	if (!pass) {
		send_error(data, "getpwnam: Can't get information about "
			   "user: %s [before chroot]", data->auth.user);
		goto _error;
	}

	ok = authenticate(data, pass->pw_dir);
	if (ok < 0) {
		goto _error;
	}

	if (ok) {
		debug("Sending AUTH OK packet");

		if (send_empty(data, PTYPE_AUTH_OK) < 0) {
			error("Can't write AUTH OK packet to socket");
			goto _error;
		}
	} else {
		send_message(data, "Permission denied");
		goto _error;
	}

	debug("Reading command info struct from socket");

	if (read_cmd(data->sd, &data->cmd) < 0) {
		error("Can't read command info");
		goto _error;
	}

	/* Build path to new root. */

	data->root = malloc(strlen(pass->pw_dir) + 1 +
			    strlen(data->host) + 1 + 
			    strlen(data->cmd.target) + 1);
	if (!data->root) {
		send_oom_error(data);
		goto _error;
	}

	strcpy(data->root, pass->pw_dir);
	strcat(data->root, "/");
	strcat(data->root, data->host);
	strcat(data->root, "-");
	strcat(data->root, data->cmd.target);

	/* Should we just unmount all filesystems and return? */
	if (data->cmd.umount) {
		static mount_t *nul = NULL;

		debug("--umount-all requested");

		if (unmount_opts(data) < 0) {
			goto _error;
		}

		/* Return a dummy pointer (non-error). */
		return &nul;
	}

	/* Parse mount information. */

	mounts = calloc(get_ptrv_len((void **) data->cmd.opts) + 1,
			   sizeof (mount_t *));
	if (!mounts) {
		send_oom_error(data);
		goto _error;
	}

	data->mounts = mounts;

	for (optp = data->cmd.opts, mntp = mounts; *optp; optp++) {
		mount_t *mnt;

		/* Store this for error reporting. */
		strncpy(strbak, *optp, sizeof (strbak) - 1);

		mnt = mnt_parse(data, *optp, data->root);
		if (!mnt) {
			errno = 0;
			send_error(data, "Invalid mount entry: %s", strbak);

			/* Free already allocated structures. */
			for (mntp = mounts; *mntp; mntp++) {
				mnt_free(*mntp);
				*mntp = NULL;
			}

			goto _error;
		}

		*mntp++ = mnt;
	}

	sort_mounts(mounts);

	/* Merge mount information and mount filesystems. */

	for (mntp = mounts; *mntp; ) {
		mount_t *canon;

		/* Store this for error reporting. */
		strncpy(strbak, (*mntp)->point, sizeof (strbak) - 1);

		canon = add_mount(data, *mntp, pass->pw_uid, pass->pw_gid);
		mnt_free(*mntp);

		*mntp++ = canon;

		if (!canon) {
			errno = 0;
			send_error(data, "Can't mount to: %s", strbak);

			/* Free the unmerged structures. */
			for (; *mntp; mntp++) {
				mnt_free(*mntp);
				*mntp = NULL;
			}

			goto _error;
		}
	}

	return mounts;

_error:
	if (mounts) {
		free(mounts);
	}
	return NULL;
}

static void send_rc(handler_t *data,
		    int16_t rc)
{
	debug("Sending RC packet: %d", rc);

	/* Send the return code to the client. */
	if (send_int16(data, PTYPE_RC, rc) >= 0) {
		fd_set fds;

		/* Do something until the client goes away. */
		while (1) {
			FD_ZERO(&fds);
			FD_SET(data->sd, &fds);

			if (select(data->sd + 1, &fds, NULL, NULL, NULL) < 0) {
				if (errno == EINTR) {
					continue;
				} else {
					break;
				}
			}

			if (read(data->sd, data->tmp_buf, BUFFER_SIZE) <= 0) {
				break;
			}
		}
	}
}

static void handler_handle(handler_t *data)
{
	int16_t rc = -1;
	pid_t pid;
	
	if (data->cmd.is_tty) {
		int fd;

		debug("Creating command process in a pty");

		pid = fork_pty(data, &fd);

		data->in.fd = fd;
		data->out.fd = fd;
		data->err.fd = -1;
	} else {
		debug("Creating command process without a pty");

		pid = fork_sockets(data, &data->in.fd,
				   &data->out.fd, &data->err.fd);
	}

	if (pid == 0) {
		/* Child: */

		execute_command(data);

		/* Not reached. */
	}

	/* Parent: */

	if (pid > 0) {
		rc = handler_manage(data, pid);
	}

	if (rc < 0) {
		rc = INTERNAL_ERROR_CODE;
	}

	send_rc(data, rc);
}

static void release_mounts(mount_t **vec)
{
	while (*vec) {
		release_mount(*vec++);
	}
}

static handler_t *alloc_handler(void)
{
	handler_t *h;

	h = calloc(1, sizeof (handler_t));

	if (h) {
		h->sd = -1;

		h->in.req_ptype = PTYPE_IN_REQ;
		h->out.data_ptype = PTYPE_OUT_DATA;
		h->err.data_ptype = PTYPE_ERR_DATA;

		h->error = 0;
	}

	return h;
}

/**
 * Frees all resources used by a handler.
 */
static void free_handler(handler_t *data)
{
	buf_free(&data->in.buf);

	if (data->root) {
		free(data->root);
	}

	free_cmd(&data->cmd);

	if (data->sd >= 0) {
		close(data->sd);
	}

	free(data);
}

static void pid_mounts_add(pid_t pid,
			   mount_t **vec)
{
	pid_mounts_t *pm, *node;

	/* Allocate */

	pm = calloc(1, sizeof (pid_mounts_t));
	if (!pm) {
		oom_error();
		return;
	}

	pm->pid = pid;
	pm->mounts = vec;

	/* Append */

	if (pid_mounts == NULL) {
		pid_mounts = pm;
	} else {
		for (node = pid_mounts; node->next; node = node->next);
		node->next = pm;
	}
}

static void pid_mounts_del(pid_t pid)
{
	pid_mounts_t *pm = NULL, *node;

	if (!pid_mounts) {
		return;
	}

	/* Find and remove */

	if (pid_mounts->pid == pid) {
		pm = pid_mounts;
		pid_mounts = pm->next;
	} else {
		for (node = pid_mounts; node->next; node = node->next) {
			if (node->next->pid == pid) {
				pm = node->next;
				node->next = pm->next;
				break;
			}
		}
	}

	if (!pm) {
		return;
	}

	debug("Found mounts for pid %d", pid);

	/* Deallocate */

	if (pm->mounts) {
		release_mounts(pm->mounts);
		free(pm->mounts);
	}

	free(pm);
}

static void accept_conn(int srvsd)
{
	struct sockaddr_in addr;
	socklen_t len = sizeof (addr);
	int clisd;
	handler_t *data;
	mount_t **mounts = NULL;
	pid_t pid;

	clisd = accept(srvsd, (struct sockaddr *) &addr, &len);
	if (clisd < 0) {
		error("Can't accept connection");
		return;
	}

	debug("New connection");

	data = alloc_handler();
	if (!data) {
		oom_error();
		close(clisd);
		return;
	}

	data->sd = clisd;

	/* Get the IP address. */
	if (getnameinfo((struct sockaddr *) &addr, len,
			data->host, sizeof (data->host),
			NULL, 0, NI_NUMERICHOST)) {
		error("Can't get client's IP address");
		goto _error;
	}

	mounts = handler_startup(data);
	if (!mounts) {
		goto _error;
	}

	if (data->cmd.umount) {
		/* We've done all we should for the client. */

		send_rc(data, 0);

		free_handler(data);
		return;
	}

	/* Start handler process. */
	pid = fork();
	if (pid < 0) {
		error("Can't fork");
		goto _error;
	}

	if (pid == 0) {
		/* Child: */

		close(srvsd);

		handler_handle(data);

		free_handler(data);
		free(mounts);

		debug("Handler process exiting");
		exit(0);
	}

	/* Parent: */

	/* Map pid to mounts vector. */
	pid_mounts_add(pid, mounts);

	free_handler(data);
	return;

_error:
	send_rc(data, INTERNAL_ERROR_CODE);

	free_handler(data);
	if (mounts) {
		free(mounts);
	}
}

/**
 * Unmounts all filesystems that we've mounted and exits.
 */
static void clean_exit(int rc)
{
	unmount_all();

	debug("sbrshd exiting");
	exit(rc);
}

static void sig_print(int sig)
{
	debug("Signal received: %s (%d)", strsignal(sig), sig);
}

static void sig_exit(int sig)
{
	sig_print(sig);

	if (getpid() == daemon_pid) {
		clean_exit(0);
	}
	exit(0);
}

static void usage(char *progname)
{
	fprintf(stderr, "Usage: %s [-p|--port <port>]"
				 " [-d|--debug <path> [-i|--io-debug]]"
				 " [-m|--mount-bin <path>]"
				 " [-u|--umount-bin <path>]"
				 " [-t|--mount-tab <path>]"
				 " [-b|--bind-opt <options>]"
				 " [-e|--mount-expiration <minutes>]\n"
			"       %s -v|--version\n"
			"       %s -h|--help\n",
		progname, progname, progname);

	exit(0);
}

/**
 * Reads options. Prints usage and exits when necessary.
 */
static void read_args(char *progname,
		      int argc,
		      char **argv)
{
	struct option longopts[] = {
		{ "help",             no_argument,       0, 'h' },
		{ "version",          no_argument,       0, 'v' },
		{ "port",             required_argument, 0, 'p' },
		{ "debug",            required_argument, 0, 'd' },
		{ "mount-bin",        required_argument, 0, 'm' },
		{ "umount-bin",       required_argument, 0, 'u' },
		{ "mount-tab",        required_argument, 0, 't' },
		{ "bind-opt",         required_argument, 0, 'b' },
		{ "mount-expiration", required_argument, 0, 'e' },
		{ 0 }
	};

	char *exp_str = NULL;
	char *debugname = NULL;

	while (1) {
		int c;

		c = getopt_long(argc, argv, "hvp:d:m:u:t:b:e:", longopts, NULL);
		if (c < 0) {
			break;
		}

		switch (c) {
		case 'p':
			port = atoi(optarg);
			break;

		case 'd':
			debugname = optarg;
			break;

		case 'm':
			mount_cmd = optarg;
			break;

		case 'u':
			umount_cmd = optarg;
			break;

		case 't':
			mount_tab = optarg;
			break;

		case 'b':
			bind_opt = optarg;
			break;

		case 'e':
			exp_str = optarg;
			break;

		case 'v':
			fprintf(stderr, DAEMON_VERSION "\n");
			exit(0);

		case 'h':
		case '?':
		default:
			usage(progname);
		}
	}

	if (port <= 0 || port >= 65536) {
		error_err(progname, "Invalid port number: %d", port);
		exit(1);
	}

	if (exp_str) {
		int exp;

		exp = atoi(exp_str);
		if (exp <= 0) {
			error_err(progname,
				  "Invalid expiration time: %s minutes",
				  exp_str);
			exit(1);
		}

		mount_expiration = exp * 60;
	}

	if (debugname) {
		debug_file = fopen(debugname, "a");
		if (!debug_file) {
			error_err(progname, "Can't open debug log file: %s",
				  debugname);
		}
	}
}

/**
 * Parse args, open logs, create a socket, accept connections, fork handlers.
 */
int main(int argc,
	 char **argv)
{
	char *progname;
	int srvsd, yes = 1;
	struct sockaddr_in srvaddr = { 0 };
	pid_t pid;
	int timeout;
	struct timeval tv;

	progname = get_progname(argv[0]);
	openlog(progname, LOG_PID, LOG_DAEMON);
	read_args(progname, argc, argv);

	/* Create a socket, bind it and listen to it. */
	srvsd = socket(PF_INET, SOCK_STREAM, 0);
	if (srvsd < 0) {
		error_err(progname, "Can't create socket");
		return 1;
	}

	if (setsockopt(srvsd, SOL_SOCKET, SO_REUSEADDR,
		       &yes, sizeof (yes)) < 0) {
		error_err(progname, "Can't set socket option (SO_REUSEADDR)");
		return 1;
	}

	srvaddr.sin_family = AF_INET;
	srvaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	srvaddr.sin_port = htons(port);

	if (bind(srvsd, (struct sockaddr *) &srvaddr, sizeof (srvaddr)) < 0) {
		error_err(progname, "Can't bind socket to port %d", port);
		return 1;
	}

	if (listen(srvsd, BACKLOG) < 0) {
		error_err(progname, "Can't listen with socket");
		return 1;
	}

	/* Daemon mode; kill the parent, etc. */

	pid = fork();
	if (pid < 0) {
		error_err(progname, "Can't fork");
		return 1;
	}
	if (pid > 0) {
		printf("%d\n", pid);
		return 0;
	}

	setsid();
	chdir("/");
	umask(0);

	debug("sbrshd version " VERSION " running at port %d", port);

	daemon_pid = getpid();

	signal(SIGINT, sig_exit);
	signal(SIGTERM, sig_exit);
	signal(SIGHUP, sig_exit);
	signal(SIGCHLD, sig_print);

	timeout = mount_expiration / MOUNT_EXPIRATION_FREQUENCY;

	tv.tv_sec = timeout;
	tv.tv_usec = 0;

	/* Accept connections from clients. */
	while (1) {
		fd_set fds;
		int count;

		debug("Waiting for connection");

		FD_ZERO(&fds);
		FD_SET(srvsd, &fds);

		count = select(srvsd + 1, &fds, NULL, NULL, &tv);

		if (count < 0) {
			/* Select failed? */
			if (errno != EINTR) {
				error("Select failed");
				clean_exit(1);
			}

			/* We received a signal. */
			while (1) {
				pid_t pid;
				int status;

				pid = waitpid(-1, &status, WNOHANG);
				if (pid <= 0) {
					break;
				}

				/* Release mounts if found. */
				print_status(pid, status);
				pid_mounts_del(pid);
			}
		} else if (count == 1) {
			/* We received a connection. */
			accept_conn(srvsd);
		} else if (count == 0) {
			/* Time to expire mounts. */
			expire_mounts();

			tv.tv_sec = timeout;
			tv.tv_usec = 0;
		}
	}

	/* Never reached. */
	return 0;
}
