/*
 * Scratchbox Remote Shell
 * Copyright (c) 2003 Timo Savola <timo.savola@movial.fi>
 *
 * sbrsh is licensed under GPL.
 * You can read the full license at http://www.gnu.org/licenses/gpl.txt
 */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <fakeroot-net.h>
#include "daemon.h"
#include "fakeroot.h"

typedef enum { DAEMON_TO_CLIENT, CLIENT_TO_DAEMON } direction_t;

struct node_s;
typedef struct node_s {
	struct node_s *next;
	int client;
	int daemon;
} node_t;

static void append_node(handler_t *data,
			node_t **head,
			int client,
			int daemon)
{
	node_t *node, *n;

	node = calloc(1, sizeof (node_t));
	if (!node) {
		send_oom_error(data);
		exit(1);
	}

	node->client = client;
	node->daemon = daemon;

	if (*head) {
		for (n = *head; n->next; n = n->next);
		n->next = node;
	} else {
		*head = node;
	}
}

static void remove_node(node_t **head,
			node_t *node)
{
	node_t *n;

	if (*head == node) {
		*head = node->next;
	} else {
		for (n = *head; n->next != node; n = n->next);
		n->next = node->next;
	}

	debug("Removing node: client=%d, daemon=%d",
	      node->client, node->daemon);

	close(node->client);
	close(node->daemon);

	free(node);
}

static void get_device_numbers(handler_t *data)
{
	mount_t **ptr;
	int index = strlen(data->root);

	for (ptr = data->mounts; *ptr; ptr++) {
		mount_t *m = *ptr;
		char *path;
		struct stat buf;

		if (m->type != MTYPE_NFS || m->point_dev > 0LL) {
			continue;
		}

		path = &m->point[index];

		if (stat(path, &buf) < 0) {
			send_error(data, "Can't stat [%s]%s",
				   data->root, path);
			exit(1);
		}

		debug("Device number of [%s]%s is %lld",
		      data->root, path, buf.st_dev);

		m->point_dev = buf.st_dev;
	}
}

static void translate_dev(handler_t *data,
			  struct fakestat *st,
			  direction_t direction)
{
	mount_t **ptr;
	dev_t old = ntohll(st->dev);

	for (ptr = data->mounts; *ptr; ptr++) {
		mount_t *m = *ptr;
		dev_t new;

		if (direction == DAEMON_TO_CLIENT) {
			if (old != m->device_dev) {
				continue;
			}
			new = m->point_dev;
		} else {
			if (old != m->point_dev) {
				continue;
			}
			new = m->device_dev;
		}

		debug("Translating device number %lld to %lld", old, new);

		st->dev = htonll(new);
		return;
	}

	debug("No match for device number %lld", old);
}

static int copy_msg(handler_t *data,
		    int source,
		    int destination,
		    direction_t direction)
{
	struct fake_msg buf;
	int len;

	debug(direction == DAEMON_TO_CLIENT ?
	      "Copying message from daemon=%d to client=%d" :
	      "Copying message from client=%d to daemon=%d",
	      source, destination);

	len = read(source, &buf, sizeof (buf));
	if (len < 0) {
		send_error(data, "Can't read from socket");
		exit(1);
	}

	if (len == 0) {
		return -1;
	}

	translate_dev(data, &buf.st, direction);

	if (write(destination, &buf, sizeof (buf)) < 0) {
		send_error(data, "Can't write to socket");
		exit(1);
	}

	return 0;
}

static int get_daemon(handler_t *data,
		      struct sockaddr_in *addr)
{
	int sd;

	sd = socket(PF_INET, SOCK_STREAM, 0);
	if (sd < 0) {
		send_error(data, "Can't create socket");
		exit(1);
	}

	if (connect(sd, (struct sockaddr *) addr,
		    sizeof (struct sockaddr_in)) < 0) {
		if (errno == EINTR) {
			debug("Connect interrupted");
			exit(0);
		}

		send_error(data, "Connect failed");
		exit(1);
	}

	debug("Connected to daemon=%d", sd);

	return sd;
}

static int get_client(handler_t *data,
		      int listen_sd)
{
	struct sockaddr_in addr;
	socklen_t len;
	int sd;

	len = sizeof (addr);
	sd = accept(listen_sd, (struct sockaddr *) &addr, &len);
	if (sd < 0) {
		if (errno == EINTR) {
			debug("Accept interrupted");
			exit(0);
		}

		send_error(data, "Accept failed");
		exit(1);
	}

	debug("Connection from client=%d", sd);

	return sd;
}

static void do_relay(handler_t *data,
		     int listen_sd,
		     int port)
{
	struct hostent *host;
	struct sockaddr_in addr = { 0 };
	node_t *sd_head = NULL;
	fd_set fds;

	host = gethostbyname(data->host);
	if (!host) {
		send_error(data, "Can't resolve host: %s", data->host);
		exit(1);
	}

	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = *(unsigned int *) host->h_addr;
	addr.sin_port = htons(port);

	debug("Fakeroot daemon assumed to be at %s:%d", data->host, port);

	while (1) {
		int maxfd, count;
		node_t *n, *next;

		FD_ZERO(&fds);

		FD_SET(listen_sd, &fds);
		maxfd = listen_sd;

		for (n = sd_head, count = 0; n; n = n->next, count++) {
			FD_SET(n->daemon, &fds);
			FD_SET(n->client, &fds);
			maxfd = MAX(maxfd, n->daemon);
			maxfd = MAX(maxfd, n->client);
		}

		debug("Selecting (%d nodes)", count);

		if (select(maxfd + 1, &fds, NULL, NULL, NULL) < 0) {
			if (errno == EINTR) {
				debug("Select interrupted");
				exit(0);
			}

			send_error(data, "Select failed");
			exit(1);
		}

		for (n = sd_head; n; n = next) {
			next = n->next;

			if (FD_ISSET(n->daemon, &fds) &&
			    copy_msg(data, n->daemon, n->client,
				     DAEMON_TO_CLIENT) < 0) {
				remove_node(&sd_head, n);
				continue;
			}

			if (FD_ISSET(n->client, &fds) &&
			    copy_msg(data, n->client, n->daemon,
				     CLIENT_TO_DAEMON) < 0) {
				remove_node(&sd_head, n);
			}
		}

		if (FD_ISSET(listen_sd, &fds)) {
			append_node(data, &sd_head,
				    get_client(data, listen_sd),
				    get_daemon(data, &addr));
		}
	}
}

static int create_relay(handler_t *data,
			int faked_port)
{
	int listen_sd, yes = 1;
	struct sockaddr_in addr;
	socklen_t len;
	pid_t pid;

	listen_sd = socket(PF_INET, SOCK_STREAM, 0);
	if (listen_sd < 0) {
		send_error(data, "Can't create socket");
		return -1;
	}

	if (setsockopt(listen_sd, SOL_SOCKET, SO_REUSEADDR,
		       &yes, sizeof (yes)) < 0) {
		send_error(data, "Can't set socket option: SO_REUSEADDR");
		goto _error;
	}

	if (listen(listen_sd, BACKLOG) < 0) {
		send_error(data, "Can't listen to socket");
		goto _error;
	}

	len = sizeof (addr);
	if (getsockname(listen_sd, (struct sockaddr *) &addr, &len) < 0) {
		send_error(data, "Can't get socket name");
		goto _error;
	}

	pid = fork();
	if (pid < 0) {
		send_error(data, "Can't fork");
		goto _error;
	}

	if (pid > 0) {
		/* Parent: */
		close(listen_sd);
		return ntohs(addr.sin_port);
	}

	/* Child: */
	get_device_numbers(data);
	do_relay(data, listen_sd, faked_port);
	exit(0);

_error:
	close(listen_sd);
	return -1;
}

int fakeroot_relay(handler_t *data)
{
	char *str;
	int port;

	str = getenv(FAKEROOTKEY_ENV);
	if (!str || !*str) {
		return 0;
	}

	port = atoi(str);
	if (port <= 0 || port >= 65536) {
		send_error(data, "Invalid " FAKEROOTKEY_ENV ": %s", str);
		return -1;
	}

	port = create_relay(data, port);
	if (port < 0) {
		return -1;
	}

	str = malloc(strlen(FAKEROOTKEY_ENV) + 1 + 5 + 1);
	if (!str) {
		send_oom_error(data);
		return -1;
	}

	sprintf(str, FAKEROOTKEY_ENV "=%d", port);
	putenv(str);

	debug("Relaying fakeroot messages at port %d", port);
	return 0;
}
